// LZSS -    (   )
//
// icq:133575351    email:lawnmower-man@mail.ru
//
// :
// DATA 1[7c] 			-  		c=1-128 
// REF  01[6c] [x8]		-  	=3- 66 ,  1-256 
// REF  00[6c] [x16]	- 			c=3- 66 ,  257-65792 

#include<stdio.h>
#include<memory.h>

#define OFFSET_MIN1 1				//!<   
#define OFFSET_MAX1 256				//!<   
#define OFFSET_MIN2 257				//!<   
#define OFFSET_MAX2 (257+65792)		//!<   
#define COUNT_MIN   3				//!<   
#define COUNT_MAX   (63+3)			//!<    
#define BUFFER_SIZE (65536*2)		//!<   
#define DATA_MIN	1				//!<   
#define DATA_MAX	128				//!<   

class LZSS{
	int data_count;					//!<   
	int math_count;					//!<   
	int math_offset;				//!<   
	int index1,index2;				//!<    (,)
	char buffer[BUFFER_SIZE];		//!<  
//    
	inline int GetSize();					//!<     
	inline int IncIndex(int,int step=1);	//!<     
	inline int DecIndex(int,int step=1);	//!<     
//    
	void FlushData(FILE*);					//!<     
	void FlushMath(FILE*);					//!<     
	int  FindMath(int*);					//!<       
// 
	int GetFileSize(FILE*);					//!<   
public:
	void Pack(FILE*,FILE*);					//!< 
	void Depack(FILE*,FILE*);				//!< 
	static void Pack(char*,char*);
	static void Depack(char*,char*);
};


int LZSS::GetFileSize(FILE *file)
{
	int size;
	fseek(file,0,SEEK_END);
	size=ftell(file);
	fseek(file,0,SEEK_SET);
	return size;
}


void LZSS::FlushData(FILE* file)
{
	if(data_count<DATA_MIN) return;
	fputc((data_count-DATA_MIN)|0x80,file);
	for(int n=0; n<data_count; n++) {
		fputc(buffer[index1],file);
		index1=IncIndex(index1);
	}
	data_count=0;
}

void LZSS::FlushMath(FILE *file)
{
	if(math_count<COUNT_MIN) return;
	if(math_offset>OFFSET_MAX1) {
		math_offset-=OFFSET_MIN2;
		fputc((math_count-COUNT_MIN),file);
		fputc(math_offset&0xFF,file);
		fputc(math_offset>>8,file);
	}else{
		math_offset-=OFFSET_MIN1;
		fputc((math_count-COUNT_MIN)|0x40,file);
		fputc(math_offset,file);
	}
	index1=IncIndex(index1,math_count);
}


int LZSS::GetSize()
{
	return (index2-index1+BUFFER_SIZE)%BUFFER_SIZE;
}


int LZSS::IncIndex(int index,int step)
{
	return (index+step)%BUFFER_SIZE;
}


int LZSS::DecIndex(int index,int step)
{
	return (index+BUFFER_SIZE-step)%BUFFER_SIZE;
}

//! \retval offset   
//! \result  
int LZSS::FindMath(int *offset)
{
	int m,n,size,delta,count;
	if(math_count) {	//      
		if(buffer[index2]==buffer[DecIndex(index2,math_offset)]) return math_count+1;
	}
	size=GetSize()+1;
	count=0;
	for(n=1; n<OFFSET_MAX2; n++) {
		delta=size; if(delta>COUNT_MAX) delta=COUNT_MAX;
		for(m=0; m<delta; m++) {
			if(buffer[DecIndex(index2,m)]!=buffer[DecIndex(index2,n+m)]) break;
		}
		if(m>=COUNT_MIN && m>count) count=m,*offset=n;
	}
	return count;
}


void LZSS::Pack(FILE *input,FILE *output)
{
	int math_count,math_offset,sym,filesize,count;
	index1=index2=data_count=math_count=count=0;
	memset(buffer,0,sizeof(buffer));
	filesize=GetFileSize(input);
	for(;;) {
		this->math_count=math_count;
		this->math_offset=math_offset;
		sym=fgetc(input); if(sym==EOF) break;
		count++; printf("-%d   \r",filesize-count);
		buffer[index2]=sym;
		math_count=FindMath(&math_offset);
		index2=IncIndex(index2);
		if(this->math_count && math_count<=this->math_count) {
			data_count=GetSize()-this->math_count-1;
			FlushData(output);
			FlushMath(output);
			math_count=0;
		}
		data_count=GetSize();
		if(data_count==DATA_MAX) {
			FlushData(output);
			math_count=0;
		}
	}
	data_count=GetSize()-this->math_count;
	FlushData(output);
	FlushMath(output);
}


void LZSS::Depack(FILE *input,FILE *output)
{
	int sym,n;
	index1=index2=data_count=0;
	memset(buffer,0,sizeof(buffer));
	for(;;) {
		sym=fgetc(input); if(sym==EOF) break;
		if(data_count) {
			data_count--;
			buffer[index2]=sym;
			index2=IncIndex(index2);
			fputc(sym,output);
			continue;
		}
		if(sym&0x80) {data_count=(sym&0x7F)+DATA_MIN;continue;}
		math_count=(sym&0x3F)+COUNT_MIN; 
		math_offset=getc(input); if(math_offset==EOF) break;
		if(sym&0x40) math_offset++;
		else {
			sym=getc(input); if(sym==EOF) break;
			math_offset|=sym<<8;
			math_offset+=OFFSET_MIN2;
		}
		for(n=0; n<math_count; n++) {
			sym=buffer[DecIndex(index2,math_offset)]; buffer[index2]=sym;
			index2=IncIndex(index2);
			fputc(sym,output);
		}
	}
}


void LZSS::Pack(char *inname,char *outname)
{
	LZSS obj;
	printf("Pack: %s\n",inname);
	FILE *input=fopen(inname,"rb");
	FILE *output=fopen(outname,"wb");
	if(!input || !output) {printf("IO error\n");return;}
	obj.Pack(input,output);
	fclose(input);
	fclose(output);
}


void LZSS::Depack(char *inname,char *outname)
{
	LZSS obj;
	printf("Depack: %s\n",inname);
	FILE *input=fopen(inname,"rb");
	FILE *output=fopen(outname,"wb");
	if(!input || !output) {printf("IO error\n");return;}
	obj.Depack(input,output);
	fclose(input);
	fclose(output);
}


int main(int count,char **list)
{
	if(count!=4) {
		printf("%s -e infile outfile\n",list[0]);
		printf("%s -d infile outfile\n",list[0]);
		return 1;
	}
	if(list[1][0]!='-' || list[1][1]!='e' && list[1][1]!='d') {
		printf("Syntax error\n\a");
		return 2;
	}
	switch(list[1][1]) {
	case 'e': LZSS::Pack(list[2],list[3]); break;	// 
	case 'd': LZSS::Depack(list[2],list[3]); break;	// 
	}
	return 0;
}
